#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "mpi.h"
#define DEF_N 64
#define MTAG 99

int rank, nprocs;

void printArray(char* lab, int A[], int n) {
  int i;
  printf("%d: %s: ", rank, lab);  
  for (i=0; i<n; i++) {
    printf(" %2d", A[i]);
  }
  printf("\n");
} //printArray()  

void genArray(int A[], int n, int s, int range) {
  int i;
  srand(s);
  for (i=0; i < n; i++)
    A[i] = rand() % range;
} //genArray()   

// return local length of n on process r of p process
int G2L(int n, int r, int p) {
  int nl = (n + p - 1) / p;
  if (nl > n - r*nl) nl = n - r*nl;
  if (nl < 0) nl = 0;
  return(nl);
}

int intComp(const void *a, const void *b) {
  int ai = *((int *)a), bi = *((int *)b);
  if (ai < bi)
    return -1;
  else if (ai > bi)
    return +1;
  else
    return 0;
}

// check global array with local portion As[0:nS]  is sorted version of A[0:n]
void checkResult(int *A, int n, int *As, int nS) {
  int ng, nSg;
  int i, nErr=0, nErrg;
  long check = 0, checkS = 0;
  MPI_Allreduce(&n, &ng, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
  MPI_Allreduce(&nS, &nSg, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
  if (ng != nSg && rank == 0)
    printf("error: global input and sorted array lengths mismatch: %d!=%d\n", 
	   ng, nSg);
  for (i = 0; i < n; i++)
    check += A[i];
  MPI_Allreduce(MPI_IN_PLACE, &check, 1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
  for (i = 0; i < nS; i++)
    checkS += As[i];
  MPI_Allreduce(MPI_IN_PLACE, &checkS, 1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
  if (check != checkS && rank==0)
    printf("error: checksum differ in global sorted & original array %ld\n", 
	   checkS - check);    

  for (i = 1; i < nS; i++)
    nErr += (As[i-1] > As[i]);
  MPI_Allreduce(&nErr, &nErrg, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);  
  if (nErrg > 0 && rank==0)
    printf("error: %d intra-bucket elements are out-of-order\n", nErrg);

  //### improve this code to better handle the case of nS==0
  nErr = 0; 
  if (rank+1 < nprocs) {
    int topElt = nS>0? As[nS-1]: 0; //this is a kludge!
    MPI_Send(&topElt, 1, MPI_INT, rank+1, 0, MPI_COMM_WORLD);
  }
  if (rank-1 >= 0) {
    int AsLeft;
    MPI_Recv(&AsLeft, 1, MPI_INT, rank-1, 0, MPI_COMM_WORLD, 
	     MPI_STATUS_IGNORE);
    nErr = (nS > 0 && AsLeft > As[0]); //another kludge!
  }
  MPI_Allreduce(&nErr, &nErrg, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);  
  if (nErrg > 0 && rank==0)
    printf("error: %d inter-bucket elements are out-of-order\n", nErrg);
} //checkResult()


int main(int argc, char *argv[]) {
  int *A; int n, n_loc, rangeA; 
  int *AbL, *bLsize, *bLoffs; //nproc local buckets
  int *AbC, nbC=0, *bCsizeR, *bCoffsR; //this process' combined bucket 
  int i, b;
  double t1, t2;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  n = (argc > 1)? atoi(argv[1]): 1;
  n_loc = G2L(n, rank, nprocs);
  rangeA = (n+1)/2;
  A = malloc(n_loc*sizeof(int));
  genArray(A, n_loc, rank*100, rangeA);  

  MPI_Barrier(MPI_COMM_WORLD);
  t1 = MPI_Wtime();

  // split A[0:n_loc-1] into nproc local buckets, 
  // bucket b starts at bLoffs[b] and has length bLsize[b]
  AbL = malloc(n_loc*sizeof(int));
  bLsize = calloc(nprocs, sizeof(int));
  bLoffs = calloc(nprocs, sizeof(int));
  for (i = 0; i < n_loc; i++) // work out how many elements each local bucket 
    bLsize[A[i] * nprocs / rangeA]++;
  bLoffs[0] = 0; 
  for (b = 1; b < nprocs; b++) // get starting offsets of local buckets
    bLoffs[b] = bLoffs[b-1] + bLsize[b-1];
  for (b = 0; b < nprocs; b++) 
    bLsize[b] = 0;
  for (i = 0; i < n_loc; i++) { // put each element into a bucket
    int b = A[i] * nprocs / rangeA;
    AbL[bLoffs[b] + bLsize[b]] = A[i];
    bLsize[b]++;
  }
  for (b = 0; b < nprocs-1; b++) //some sanity checks
    assert (bLoffs[b] + bLsize[b] == bLoffs[b+1]);
  assert (bLoffs[nprocs-1] + bLsize[nprocs-1] == n_loc);

  bCsizeR = malloc(nprocs * sizeof(int));
  bCoffsR = malloc(nprocs * sizeof(int));

  //### add code so that bCsizeR[b] has the bLsize[rank] from process b


  //printArray("bCs", bCsizeR, nprocs); //uncomment for debugging

  //### add code to compute bCoffsR[0:nprocs-1 and nbC from bCsizeR[0:nprocs-1]


  // now we have nbC, we can allocate the combined bucket for this process
  AbC = calloc(nbC==0? 1: nbC, sizeof(int));

  //### finally, add code to put required data into combined bucket AbC 

  
  //printArray("AbC", AbC, nbC); //uncomment for debugging
  // perform local sort on combined bucket
  qsort((void *)AbC, nbC, sizeof(int), intComp);

  printf("%d: bucket %d elts in %d..%d\n", rank, nbC, AbC[0], AbC[nbC==0? 0: nbC-1]);

  MPI_Barrier(MPI_COMM_WORLD);
  t2 = MPI_Wtime() - t1;

  if (rank==0)
    printf("sort %d elements in range 0..%d across %d procs in time %.2es\n", 
	   n, rangeA-1, nprocs, t2);
  checkResult(A, n_loc, AbC, nbC);
 	   		 
  free(AbC); free(bCoffsR); free(bCsizeR);
  free(AbL); free(bLsize); free(bLoffs);
  free(A); 
  MPI_Finalize();
  return 0;
} //main()
